#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from typing import Iterable, Callable, Iterator, Tuple, Optional, NamedTuple
import numpy as np

from pinecone.protos import core_pb2
from pinecone.utils import load_numpy
from pinecone.utils.sentry import sentry_decorator as sentry
from pinecone.utils.progressbar import ProgressBar
from .api_controller import ControllerAPI
from .api_router import RouterAPI
from .constants import Config
from .grpc import GRPCClient

__all__ = ["connect", "Connection", "Cursor", "IndexResult", "DeleteResult", "QueryResult", "InfoResult", "FetchResult"]


class IndexResult(NamedTuple):
    """Result of an index request."""

    id: str


class DeleteResult(NamedTuple):
    """Result of a delete request."""

    id: str


class QueryResult(NamedTuple):
    """Result of a query request."""

    ids: Iterable[str]
    scores: Iterable[float]
    data: np.ndarray = None


class InfoResult(NamedTuple):
    """Result of an info request"""

    index_size: str


class FetchResult(NamedTuple):
    """Result of a fetch request"""

    id: str
    vector: np.ndarray


class CursorConfig(NamedTuple):
    """Cursor configurations.


    :param grpc_client: the cursor's gRPC client.
    :type grpc_client: :class:`GRPCClient`
    :param create_request_fn: the function to create gRPC requests given the items.
    :type create_request_fn: Callable[[Iterable], "core_pb2.Request"]
    :param parse_response_fn: the function to parse gRPC response
    :type parse_response_fn: Optional[Callable[["core_pb2.Request"], Iterable]], optional
    :param unary: send unary-unary gRPC requests. Defaults to `False`.
    :type unary: bool, optional
    :param disable_progress_bar: disable progress bar. Defaults to `False`.
    :type disable_progress_bar: bool
    """

    grpc_client: GRPCClient
    create_request_fn: Callable[[Iterable], "core_pb2.Request"]
    parse_response_fn: Optional[Callable[["core_pb2.Request"], Iterable]] = None
    unary: Optional[bool] = False
    disable_progress_bar: bool = False


class Cursor:
    def __init__(
        self,
        config: CursorConfig,
        items: Iterable[np.ndarray],
        batch_size: Optional[int] = 100,
    ):
        """A cursor batch-processes requests sent to a service.

        :param config: cursor configurations
        :type config: :class:`CursorConfig`
        :param items: Items to process
        :type items: Iterable[np.ndarray]
        :param batch_size: the number of items to batch-process, defaults to 100
        :type batch_size: int, optional
        """
        self.batch_size = batch_size or 100
        self._item_iterator = iter(items)
        self._unary = config.unary
        self._disable_progress_bar = config.disable_progress_bar
        self._client = config.grpc_client
        self._create_reqeust_fn = config.create_request_fn
        self._parse_response_fn = config.parse_response_fn or self.identity

    @sentry
    def take(self, size: int):
        """Processes a given number of items and returns the results.

        :param size: the number of items to process
        """
        result = []
        if self._disable_progress_bar:
            result = list(self._fetch_unbuffered(size=size))
        else:
            result = list(ProgressBar.iter(self._fetch_unbuffered(size=size)))
        return result

    @sentry
    def collect(self):
        """Processes all of the items and returns the results."""
        result = []
        if self._disable_progress_bar:
            result = list(self._fetch_unbuffered())
        else:
            result = list(ProgressBar.iter(self._fetch_unbuffered()))
        return result

    @sentry
    def stream(self) -> Iterator:
        """Processes all of the items, returns the results one batch at a time."""
        ret_iter = None
        if self._disable_progress_bar:
            ret_iter = self._fetch_unbuffered()
        else:
            ret_iter = ProgressBar.iter(self._fetch_unbuffered())
        return ret_iter

    def __iter__(self):
        raise RuntimeError(
            "To iterate through the results, "
            "consider using one of the followoing methods: `take()`, `collect()`, `stream()`."
        )

    @classmethod
    def identity(cls, entity):
        """An identity function."""
        return entity

    def _fetch_unbuffered(self, size: int = None) -> Iterator:
        """Creates a generator to fetch results.

        :param size: how many items to fetch.
        """

        def parse_response(response):
            if response.status.code == core_pb2.Status.StatusCode.Value("ERROR"):
                exception = bool(response.status.details) and response.status.details[0].exception
                error_msg = exception or response.status.description or "Unknown error when fetching results."
                raise RuntimeError(error_msg)

            for elem in self._parse_response_fn(response):
                yield elem

        request_generator = self._generate_requests(size)
        if self._unary:
            for request in request_generator:
                response_ = self._client.send_request(request)
                for elem in parse_response(response_):
                    yield elem
        else:
            for response_ in self._client.stream_requests(request_generator):
                for elem in parse_response(response_):
                    yield elem

    def _generate_requests(self, size: int = None) -> Iterator["core_pb2.Request"]:
        """Generates protobuf requests.

        This method consumes the item iterator in batches,
        then creates requests that each wraps the batch of items.

        :param size: how many items to consume. If `None`, th method exhausts the item iterator.
        """
        batch = []
        for ii, item in enumerate(self._item_iterator):
            batch.append(item)
            if len(batch) >= self.batch_size or (size and len(batch) >= size):
                yield self._create_reqeust_fn(batch)
                batch = []
            if not size:
                # consume everything
                continue
            elif ii >= size - 1:
                break
        if batch:
            yield self._create_reqeust_fn(batch)


class Connection:
    """A connection manages communication with a service.

    You can only interact with a service (e.g. upsert, query)
    after a connection to the service has been established.

    Namespaces partition the items in an index. When you read from or write to a namespace in an index, you will
    only access items in that particular namespace.
    For instance, two namespaces can contain the items with the same ids but different values.
    Use namespaces when you want to use the same preprocessors and postprocessors for separate datasets.
    For example, if you are building a movie recommender system, then you could use namespaces to
    separate the recommendations by genre.
    """

    def __init__(
        self,
        host: str,
        port: int,
        api_key: str = None,
        service_name: str = None,
        timeout: int = 0,
        response_timeout: int = 20,
        traceroute: bool = False,
    ):
        """
        :param host: service host
        :type host: str
        :param port: service port
        :type port: int
        :param api_key: user API key, defaults to None
        :type api_key: str, optional
        :param service_name: name of the service, defaults to None
        :type service_name: str, optional
        :param timeout: defaults to 0. Timeout to retry connection if gRPC is unavailable. 0 is no retry.
        :type timeout: int, optional
        :param response_timeout: defaults to 20 seconds. Fail if gateway doesn't receive response within timeout.
        :type response_timeout: int, optional
        :param traceroute: Whether to send receipts back to the gateway from each stage of the graph
        :type traceroute: bool, optional
        """

        self.host = host
        self.port = port
        self.api_key = api_key
        self.service_name = service_name
        self.grpc = GRPCClient(
            host=host,
            port=port,
            api_key=api_key,
            service_name=service_name,
            conn_timeout=timeout,
            timeout=response_timeout,
            traceroute=traceroute,
        )

    @sentry
    def upsert(
        self,
        items: Iterable[Tuple[str, np.ndarray]],
        namespace: str = None,
        batch_size: int = None,
        cursor_config: dict = None,
    ) -> Cursor:
        """Inserts or updates items.

        An item is an `(id, vector)` tuple.

        Insert an item into the index, or update by item id if the item's id already exists.

        :param items: tuples of the form (id, numpy.ndarray). Id's length is limited to 64 characters.
        :type items: Iterable[Tuple[str, np.ndarray]]
        :param namespace: a partition in an index. Use default namespace when not specified.
        :type namespace: str, optional
        :param batch_size: overrides default batch size
        :type batch_size: int, optional
        :param cursor_config: additional cursor configs. See :class:`CursorConfig`.
        :type cursor_config: dict, optional
        :return: :class:`Cursor`
        """

        def _create_index_request(item_batch: Iterable[Tuple[str, np.ndarray]]) -> "core_pb2.Request":
            """Creates an index request using all of the items."""
            path = "write"
            ids_buffer, vectors_buffer = list(zip(*item_batch))
            ids_buffer = list(map(str, ids_buffer))
            return self.grpc.get_index_request(ids=ids_buffer, data=vectors_buffer, namespace=namespace, path=path)

        def _parse_index_response(response):
            return [IndexResult(id=_id) for _id in response.index.ids]

        config = CursorConfig(
            grpc_client=self.grpc,
            create_request_fn=_create_index_request,
            parse_response_fn=_parse_index_response,
            **(cursor_config or {})
        )
        return Cursor(config=config, items=items, batch_size=batch_size)

    @sentry
    def unary_upsert(self, item: Tuple[str, np.ndarray], namespace: str = None) -> Optional[IndexResult]:
        """Inserts or updates an item.

        This is the unary version of :py:meth:`upsert`, which upserts one item at a time.
        """
        result = self.upsert(
            items=[item], namespace=namespace, cursor_config={"unary": True, "disable_progress_bar": True}
        ).collect()
        return result[0] if result else None

    @sentry
    def delete(
        self, ids: Iterable[str], namespace: str = None, batch_size: int = None, cursor_config: dict = None
    ) -> Cursor:
        """Deletes items by their ids.

        :param ids: ids of items
        :type ids: Iterable[str]
        :param namespace: a partition in an index. Use default namespace when not specified.
        :type namespace: str, optional
        :param batch_size: overrides default batch size
        :type batch_size: int, optional
        :param cursor_config: additional cursor configs. See :class:`CursorConfig`.
        :type cursor_config: dict, optional
        :return: :class:`Cursor`
        """

        def _create_delete_request(id_batch: Iterable[str]) -> "core_pb2.Request":
            """Generates a delete request."""
            path = "write"
            return self.grpc.get_delete_request(ids=id_batch, namespace=namespace, path=path)

        def _parse_delete_response(response):
            return [DeleteResult(id=_id) for _id in response.delete.ids]

        config = CursorConfig(
            grpc_client=self.grpc,
            create_request_fn=_create_delete_request,
            parse_response_fn=_parse_delete_response,
            **(cursor_config or {})
        )
        return Cursor(config=config, items=ids, batch_size=batch_size)

    @sentry
    def unary_delete(self, id: str, namespace: str = None) -> Optional[DeleteResult]:
        """Deletes an item by its id.

        This is the unary version of :py:meth:`delete`, which deletes one item at a time.
        """
        result = self.delete(
            ids=[id], namespace=namespace, cursor_config={"unary": True, "disable_progress_bar": True}
        ).collect()
        return result[0] if result else None

    @sentry
    def fetch(
        self, ids: Iterable[str], namespace: str = None, batch_size: int = None, cursor_config: dict = None
    ) -> Cursor:
        """Fetches items by their ids.

        :param ids: ids of items
        :type ids: Iterable[str]
        :param namespace: a partition in an index. Use default namespace when not specified.
        :type namespace: str, optional
        :param batch_size: overrides default batch size
        :type batch_size: int, optional
        :param cursor_config: additional cursor configs. See :class:`CursorConfig`.
        :type cursor_config: dict, optional
        :return: :class:`Cursor`
        """

        def _create_fetch_request(id_batch: Iterable[str]) -> "core_pb2.Request":
            """Generates a delete request."""
            path = "read"
            return self.grpc.get_fetch_request(ids=id_batch, namespace=namespace, path=path)

        def _parse_fetch_response(response):
            return [
                FetchResult(id=_id, vector=load_numpy(vector))
                for _id, vector in zip(response.fetch.ids, response.fetch.vectors)
            ]

        config = CursorConfig(
            grpc_client=self.grpc,
            create_request_fn=_create_fetch_request,
            parse_response_fn=_parse_fetch_response,
            **(cursor_config or {})
        )
        return Cursor(config=config, items=ids, batch_size=batch_size)

    @sentry
    def unary_fetch(self, id: str, namespace: str = None) -> Optional[FetchResult]:
        """Fetches an item by its id.

        This is the unary version of :py:meth:`fetch`, which fetches one item at a time.
        """
        result = self.fetch(
            ids=[id], namespace=namespace, cursor_config={"unary": True, "disable_progress_bar": True}
        ).collect()
        return result[0] if result else None

    @sentry
    def query(
        self,
        queries: Iterable[np.ndarray],
        namespace: str = None,
        top_k: int = 10,
        batch_size: int = None,
        include_data: bool = False,
        cursor_config: dict = None,
    ) -> Cursor:
        """Sends queries to the index and returns the top results ordered by their scores.

        :param queries: queries
        :type queries: Iterable[np.ndarray]
        :param namespace: a partition in an index. Use default namespace when not specified.
        :type namespace: str, optional
        :param top_k: defaults to 10, the number of top query results to return for each query,
            ordered by their scores
        :type top_k: int, optional
        :param batch_size: overrides default batch size
        :type batch_size: int, optional
        :param include_data: whether to return data associated with the query results, defaults to False
        :type include_data: bool, optional
        :param cursor_config: additional cursor configs. See :class:`CursorConfig`.
        :type cursor_config: dict, optional
        :return: :class:`Cursor`
        """

        def _create_query_request(query_batch: Iterable[np.ndarray]) -> "core_pb2.Request":
            """Generates a query request."""
            path = "read"
            return self.grpc.get_query_request(
                data=query_batch, top_k=top_k, include_data=include_data, namespace=namespace, path=path
            )

        def _parse_query_response(response):
            """Parses a query response"""
            return [
                QueryResult(
                    ids=matches.ids,
                    scores=matches.scores,
                    data=load_numpy(matches.data) if response.query.include_data else None,
                )
                for matches in response.query.matches
            ]

        config = CursorConfig(
            grpc_client=self.grpc,
            create_request_fn=_create_query_request,
            parse_response_fn=_parse_query_response,
            **(cursor_config or {})
        )
        return Cursor(config=config, items=queries, batch_size=batch_size)

    @sentry
    def unary_query(
        self, query: np.ndarray, namespace: str = None, top_k: int = 10, include_data: bool = False
    ) -> Optional[QueryResult]:
        """Sends a query.

        This is the unary version of :py:meth:`query`, which sends one query at a time.
        """
        result = self.query(
            queries=[query],
            namespace=namespace,
            top_k=top_k,
            include_data=include_data,
            cursor_config={"unary": True, "disable_progress_bar": True},
        ).collect()
        return result[0] if result else None

    @sentry
    def info(self, namespace: str = None) -> Optional[InfoResult]:
        """Returns information of the index."""

        def _create_info_request(*args) -> "core_pb2.Request":
            path = "write"
            return self.grpc.get_info_request(namespace=namespace, path=path)

        def _parse_info_response(response):
            return [InfoResult(index_size=response.info.index_size)]

        cursor_config = CursorConfig(
            grpc_client=self.grpc,
            create_request_fn=_create_info_request,
            parse_response_fn=_parse_info_response,
            unary=True,
            disable_progress_bar=True,
        )
        result = Cursor(config=cursor_config, items=[1]).collect()
        return result[0] if result else None

    @sentry
    def close(self):
        """Closes the connection."""

        self.grpc.channel.close()

    def _set_secure(self, is_secure: bool):
        """Changes whether the grpc client is secure or not."""

        self.grpc.secure = is_secure


@sentry
def connect(
    service_name: str = None,
    router_name: str = None,
    timeout: int = 0,
    response_timeout: int = 20,
    traceroute: bool = False,
) -> Connection:
    """Connects to a live service/router and returns an instance of :class:`Connection`.

    :param service_name: name of a service.
    :param router_name: name of a router.
    :type service_name: str
    :type router_name: str
    :param timeout: gRPC connection timeout, defaults to 0
    :type timeout: int, optional
    :param timeout: defaults to 0. Timeout to retry connection if gRPC is unavailable. 0 is no retry.
    :type timeout: int, optional
    :param response_timeout: defaults to 20 seconds. Fail if gateway doesn't receive response within timeout.
    :type response_timeout: int, optional
    :param traceroute: Whether to send receipts back to the gateway from each stage of the graph
    :type traceroute: bool, optional
    :return: :class:`Connection`
    """
    dest_name = router_name or service_name
    if not dest_name:
        raise ValueError("Must supply service_name or router_name for connect()")
    api_cls = RouterAPI if router_name else ControllerAPI
    api = api_cls(host=Config.CONTROLLER_HOST, api_key=Config.API_KEY)
    status = api.get_status(dest_name)
    if not status.get("ready"):
        raise ConnectionError

    host = status.get("host") or Config.CONTROLLER_HOST.split("://")[1].split(":")[0]
    port = status.get("port")
    return Connection(
        host=host,
        port=port,
        api_key=Config.API_KEY,
        service_name=dest_name,
        timeout=timeout,
        response_timeout=response_timeout,
        traceroute=traceroute,
    )



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#
from pinecone.network.zmq.spec import ServletSpec, SocketType, Socket
from pinecone.network.zmq.socket_wrapper import SocketWrapper




from typing import BinaryIO
import google.protobuf.message
from pinecone.protos import core_pb2
from loguru import logger
import os
from . import load_numpy, dump_numpy
from collections import defaultdict

OFFSET_BYTES = 8
SIZE_BYTES = 4
BYTEORDER = 'big'


def read_size(log_file: BinaryIO):
    return int.from_bytes(log_file.read(SIZE_BYTES), BYTEORDER)


def read_proto(log_file: BinaryIO, size: int):
    msg_bytes = log_file.read(size)
    entry = core_pb2.Request()
    try:
        entry.ParseFromString(msg_bytes)
    except google.protobuf.message.DecodeError:
        logger.error(f"failed to load msg from WAL: {msg_bytes}")
    return entry, len(msg_bytes)


def read_entry(log_file: BinaryIO):
    size = read_size(log_file)
    entry, msg_size = read_proto(log_file, size)
    read_size(log_file)
    return entry, msg_size


def read_entry_backwards(log_file: BinaryIO) -> [core_pb2.Request, int]:
    if log_file.tell() == 0:
        return core_pb2.Request(), 0
    log_file.seek(-SIZE_BYTES, os.SEEK_CUR)
    size = read_size(log_file)
    log_file.seek(-(SIZE_BYTES + size), os.SEEK_CUR)
    entry, msg_size = read_proto(log_file, size)
    log_file.seek(-(SIZE_BYTES + size), os.SEEK_CUR)  # reset to end of previous entry
    return entry, msg_size


def write_log_entry(log_file: BinaryIO, entry: core_pb2.Request):
    log_file.seek(0, os.SEEK_END)
    entry_bytes = entry.SerializeToString()
    size = len(entry_bytes).to_bytes(SIZE_BYTES, BYTEORDER)
    log_file.write(size + entry_bytes + size)


def compact_log(log_path: str, tmp_log_path: str, end_file_offset: int):
    with open(log_path, 'rb') as log_file:
        with open(tmp_log_path, 'w+b') as tmp_log_file:
            log_file.seek(end_file_offset)
            ids_seen = defaultdict(set)
            while True:
                entry, msg_size = read_entry_backwards(log_file)
                if msg_size == 0:
                    break

                new_entry = core_pb2.Request()
                if entry.WhichOneof('body') == 'index':
                    keep_idx = []
                    for i, item_id in enumerate(entry.index.ids):
                        if item_id not in ids_seen[entry.namespace]:
                            new_entry.index.ids.append(item_id)
                            keep_idx.append(i)
                            ids_seen[entry.namespace].add(item_id)
                    data = load_numpy(entry.index.data)
                    new_entry.index.data.CopyFrom(dump_numpy(data[keep_idx, :]))
                    size = len(new_entry.index.ids)
                else:  # delete
                    for item_id in entry.delete.ids:
                        if item_id not in ids_seen[entry.namespace]:
                            new_entry.delete.ids.append(item_id)
                            ids_seen[entry.namespace].add(item_id)
                    size = len(new_entry.delete.ids)
                if size > 0:  # only write the request if it has data in it
                    write_log_entry(tmp_log_file, new_entry)



from . import Socket, SocketType
from .servlet import ZMQServlet
from pinecone.utils.constants import ZMQ_CONTROL_PORT
from pinecone.utils import get_hostname
from pinecone.protos import core_pb2

import zmq

from typing import AsyncIterable


class GatewayServlet(ZMQServlet):

    async def poll_traceroute(self) -> AsyncIterable['core_pb2.TraceRoute']:
        control_sock = self.init_socket(Socket(True, SocketType.PULL, ZMQ_CONTROL_PORT, get_hostname()))
        while True:
            yield await self.recv_traceroute(control_sock)

    async def recv_traceroute(self, sock: zmq.Socket) -> 'core_pb2.TraceRoute':
        msg = await sock.recv()
        msg_pb = core_pb2.TraceRoute()
        msg_pb.ParseFromString(msg)
        self.msg_recv += 1
        return msg_pb



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from pinecone.functions import Function

from abc import abstractmethod
from typing import Dict, Union


class Index (Function):

    """
    Generic interface for defining an index
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
    
    @abstractmethod
    def size(self) -> int:
        """
        Size of the index
        :return: the number of items in the index
        """
        raise NotImplementedError

    def get_stats(self) -> Dict[str, Union[int, float, str]]:
        return {'size': self.size()}

    




#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from typing import List
from loguru import logger
import pathlib
import shutil
import re

from pinecone.utils import hubify
from pinecone.utils.sentry import sentry_decorator as sentry
from pinecone.functions.model import HubFunction
from pinecone.functions.model.preprocessor import Preprocessor
from pinecone.functions.model.preprocessor.hub import HubPreprocessor
from pinecone.functions.model.postprocessor import Postprocessor, QueryResult
from pinecone.functions.model.postprocessor.hub import HubPostprocessor

from .constants import Config
from .api_hub import HubAPI
from .api_action import ActionAPI

__all__ = [
    "get_login_cmd",
    "list_repositories",
    "create_repository",
    "delete_repository",
    "list_repository_tags",
    "as_user_image",
    "as_repo_image",
    "ImageBuilder",
    "ImageServer",
    "preprocessor",
    "postprocessor",
    "HubFunction",
    "QueryResult",
]

# Image name consists of repository and tag. Repository and tag can only contain alphanumeric letter, period, dash, or underscore.
IMAGE_REGEX = re.compile("^([a-zA-Z0-9]|[-.]|[_.])+:([a-zA-Z0-9]|[-.]|[_.])+$")


def preprocessor(cls):
    """Decorator for wrapping a class as a Preprocessor.

    .. note::
        By default preprocessors are limited to 4GB of memory.
        If your application needs to use more memory,
        please reach out support@pinecone.io.

    Usage:

    .. code-block:: python

        from pinecone.hub import preprocessor


        @preprocessor
        class MyPreprocessor:
            def transform(self, vectors: np.ndarray) -> np.ndarray:
                return vectors

    """

    @hubify
    class PreprocessorWrapper(Preprocessor):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            # Overrides the transform function.
            self.transform = cls().transform

    return PreprocessorWrapper


def postprocessor(cls):
    """Decorator for wrapping a class as a Postporcessor.

    .. note::
        By default postprocessors are limited to 4GB of memory.
        If your application needs to use more memory,
        please reach out support@pinecone.io.

    Usage:

    .. code-block:: python

        from pinecone.hub improt postprocessor


        @postprocessor
        class MyPostprocessor:
            def transform(self, queries: np.ndarray, matches: Iterable[QueryResult]) -> Iterable[QueryResult]:
                return results

    """

    @hubify
    class PostprocessorWrapper(Postprocessor):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            # Overrides the transform function.
            self.transform = cls().transform

    return PostprocessorWrapper


def _get_hub_api():
    return HubAPI(host=Config.HUB_HOST, api_key=Config.API_KEY)


def _get_action_api():
    return ActionAPI(host=Config.CONTROLLER_HOST, api_key=Config.API_KEY)


@sentry
def get_login_cmd() -> str:
    """Returns CLI command for logging into Pinecone hub."""
    api = _get_hub_api()
    return "docker login -u {username} -p {password} {server}".format(**api.get_login())


@sentry
def list_repositories():
    """Shows all repositories."""
    api = _get_hub_api()
    return api.list_repositories()


@sentry
def create_repository(name):
    """Creates or unarchives a repository."""
    api = _get_hub_api()
    return api.create_repository(name)


@sentry
def delete_repository(name):
    """Archives a repository."""
    api = _get_hub_api()
    return api.delete_repository(name)


@sentry
def list_repository_tags(name):
    """Lists tags of a repository."""
    api = _get_hub_api()
    return api.list_repository_tags(name)


@sentry
def as_user_image(image):
    """Returns the image name prepended with the user name."""
    user_api = _get_action_api()
    username = user_api.whoami().username
    return "{username}/{image}".format(**locals())


@sentry
def as_repo_image(image):
    """Returns the complete repository image URI."""
    registry = Config.HUB_REGISTRY.split("https://")[-1].split("https://")[-1]
    user_image = as_user_image(image)
    return "{registry}/{user_image}".format(**locals())


class ImageBuilder:
    """A helper class for building Pinecone Hub models."""

    def __init__(
        self,
        image: str,
        build_path: pathlib.Path,
        model_path: pathlib.Path,
        pip: List[str] = None,
        data_paths: List[pathlib.Path] = None,
    ):
        """
        :param image: name of the docker image in the form :code:`REPOSITORY:TAG`
        :param build_path: path to which artifacts will be saved for building the docker image.
        :param model_path: path to the main model file. The model file is where Pinecone loads the inference APIs.
        :param data_paths: paths to ancillary data, such as serialized Tensorflow models.
            These files will be copied to the docker build path.
        :param pip: additional Python packages to be installed.
            In the base docker image, only *numpy* is preloaded. It is recommended that you use exact versions.
        """
        self._raise_if_invalid_image_name(image)
        self.image = image
        self.build_path = pathlib.Path(build_path)
        self.model_path = pathlib.Path(model_path)
        self.data_paths = [pathlib.Path(pp) for pp in (data_paths or [])]
        self.pip = pip or []

    def _raise_if_invalid_image_name(self, image):
        if not IMAGE_REGEX.match(image):
            raise ValueError(
                "Image name must be of the form `REPOSITORY:TAG`."
                " Repository and tag can only contain alphanumeric letter, period, or dash."
            )

    @property
    def repository(self):
        return self.image.split(":")[0]

    @property
    def tag(self):
        return self.image.split(":")[1]

    @sentry
    def package(self, exist_ok=False):
        """Creates artifacts for building a Pinecone-compatible docker image.

        The following will be created:

        - The build directory :code:`build_path` where the artifacts reside in preparation for running :code:`docker build`.
        - The main model file will be copied from :code:`model_path` to the docker build path.
        - Ancillary data files and directories specified in :code:`data_paths` will be copied to the docker build path.
        - A `Dockerfile` template for building the docker image will be created in the docker build path.

        :param exist_ok: same as `pathlib.Path.mkdir`, defaults to False.
        :type exist_ok: bool, optional
        """
        self.build_path.mkdir(parents=True, exist_ok=exist_ok)

        # Pinecone base image
        base_image = Config.BASE_IMAGE

        # Pip packages to install
        pip_str = "" if not self.pip else "RUN pip3 install --quiet --no-cache-dir {0}".format(" ".join(self.pip))

        # Copy model file
        build_model_path = pathlib.Path(self.build_path.joinpath("model.py"))
        if build_model_path.exists():
            build_model_path.unlink()
        shutil.copy(self.model_path.resolve(), build_model_path)

        # Copy data files
        copy_str_list = []
        for pp in self.data_paths or []:
            basename = pp.name
            build_data_path = pathlib.Path(self.build_path.joinpath(basename))
            if build_data_path.exists():
                if build_data_path.is_dir():
                    build_data_path.rmdir()
                else:
                    build_data_path.unlink()
            if pp.is_dir():
                shutil.copytree(pp.resolve(), build_data_path)
                copy_str_list.append("COPY {basename} ./data/{basename}/".format(**locals()))
            else:
                shutil.copy(pp.resolve(), build_data_path)
                copy_str_list.append("COPY {basename} ./data/".format(**locals()))
        copy_str = "\n".join(copy_str_list)

        # Write Dockerfile
        dockerfile_path = self.build_path.joinpath("Dockerfile")
        with pathlib.Path(dockerfile_path).open("w") as outfile:
            logger.info("Writing Dockerfile to {dockerfile_path}".format(**locals()))
            template = """# This is a Pinecone Dockerfile template.
# Required: Pinecone base image
FROM {base_image}

RUN pip3 install --quiet --upgrade pip

# Optional: additional python packages
{pip_str}

# Optional: additional paths to copy
{copy_str}

# Required: model file
COPY model.py ./model.py
"""
            outfile.write(template.format(**locals()))

        logger.info("Docker build artifacts are saved to {0}".format(self.build_path))

    @sentry
    def get_build_cmd(self) -> str:
        """Returns the docker build command."""
        build_path = self.build_path.resolve()
        image = self.image
        template = " && \\\n".join(
            [
                "pushd {build_path}",
                "docker build -t {image} .",
                "popd",
            ]
        )
        return template.format(**locals())

    @sentry
    def get_push_cmd(self) -> str:
        """Returns command to push to docker hub."""

        # Creates the repository if it doesn't already exist.
        create_repository(self.repository)

        # The repository is immutable. Verify that the tag does not already exist.
        if self.tag in list_repository_tags(self.repository):
            raise RuntimeError(
                "The tag '{0}' already exists in the '{1}' repository. "
                "Please use a new image tag and rebuild the image.".format(self.tag, self.repository)
            )

        image = self.image
        repo_image = as_repo_image(image)
        template = " && \\\n".join(
            [
                "docker tag {image} {repo_image}",
                "docker push {repo_image}",
            ]
        )
        return template.format(**locals())


class ImageServer:
    """Creates a local server for testing a HubFunction."""

    def __init__(self, image: str, image_type: str):
        """

        :param image: name of the docker image
        :type image: str
        :param image_type: one of {"preprocessor", "postprocessor"}
        :type image_type: str
        """
        self.image = image
        self.image_type = image_type
        self.hub_fn = None

    def __enter__(self):
        if self.image_type == "preprocessor":
            self.hub_fn = HubPreprocessor(image=self.image)
        elif self.image_type == "postprocessor":
            self.hub_fn = HubPostprocessor(image=self.image)
        else:
            raise RuntimeError("Image type '{0}' is not supported.".format(self.image))

        self.hub_fn.start()
        return self.hub_fn

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.hub_fn:
            self.hub_fn.stop()
            self.hub_fn.servlet.cleanup()



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

import numpy as np

from pinecone.utils import load_numpy
from pinecone.grpc import GRPCClient

from pinecone.functions.model import HubFunction


class HubPreprocessor(HubFunction):
    """Useful for local testing. Mimics a query request sent to a HubFunction hosting a Preprocessor image."""

    async def transform(self, vectors: np.ndarray, timeout: float = 30) -> np.ndarray:
        grpc = GRPCClient()
        msg = grpc.get_query_request(data=vectors, path="read")
        response = await self.handle_msg(msg, timeout)
        return load_numpy(response.query.data)



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from pinecone.functions.model import ModelFunction
from pinecone.protos import core_pb2
from pinecone.utils import load_numpy, dump_numpy

import numpy as np


class Transformer(ModelFunction):

    def setup(self):
        """
        Download and cache the model prior to serving
        :return:
        """

    def transform_vectors(self, vectors: np.ndarray) -> np.ndarray:
        """
        Transforms given set of vectors given the raw data and metadta
        :param vectors: input vectors
        :param items: metdata for each vector
        :return: numpy vectors representing tranformed vectors
        """
        raise NotImplementedError

    def transform_numpy_vectors(self, msg_vectors: 'core_pb2.NdArray') -> 'core_pb2.NdArray':
        vectors = load_numpy(msg_vectors)
        transformed_vectors = self.transform_vectors(vectors)
        return dump_numpy(transformed_vectors)

    def handle_msg(self, msg: 'core_pb2.Request') -> 'core_pb2.Request':
        req_type = msg.WhichOneof('body')
        if req_type == 'index':
            msg.index.data.CopyFrom(self.transform_numpy_vectors(msg.index.data))
        elif req_type == 'query':
            msg.query.data.CopyFrom(self.transform_numpy_vectors(msg.query.data))
        return msg



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from pinecone.protos import core_pb2
from pinecone.utils.exceptions import RestartException
from pinecone.network.zmq.spec import ServletSpec
from pinecone.network.zmq.servlet import ZMQServlet
from pinecone.utils.constants import RECV_TIMEOUT
from pinecone.network.zmq.wal import WALServlet

from loguru import logger
import asyncio
from typing import Awaitable, Callable, List


def start(servlet_spec: ServletSpec, volume_request=None, persistent_dir=None, num_replicas=None, sync_msg_handler=None,
          stats_exporter=None):
    logger.info('starting zmq servlet: %s ' % servlet_spec.function_name)

    try:
        loop = asyncio.get_event_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    if volume_request:
        servlet = WALServlet(servlet_spec, num_replicas, persistent_dir, volume_request=volume_request)
        logger.info(f"Before reload: {servlet_spec.get_stats()}")
        servlet.write_offset(servlet.replica, 0)  # replay all the messages
        servlet.load_sync(sync_msg_handler)
        logger.info(f"After reload: {servlet_spec.get_stats()}")
        loop.create_task(servlet.consume_msgs())
    else:
        servlet = ZMQServlet(servlet_spec)

    if stats_exporter:
        loop.create_task(stats_exporter())

    servlet.start_polling()
    servlet.refresh_dns()
    loop.run_forever()
    if servlet.exception is not None and type(servlet.exception) == RestartException:
        raise RestartException


def run_in_servlet(servlet_spec: ServletSpec, function: Callable[..., Awaitable], *args):
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    servlet = ZMQServlet(servlet_spec)
    result = asyncio.run(function(servlet, *args))
    loop.stop()
    return result


def run_in_wal_servlet(servlet_spec: ServletSpec, num_replicas: int, persistent_dir: str,
                       function: Callable[..., Awaitable], *args):
    try:
        loop = asyncio.get_event_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    servlet = WALServlet(servlet_spec, num_replicas, persistent_dir)
    result = asyncio.run(function(servlet, *args))
    servlet.cleanup()
    loop.stop()
    return result


async def recv_requests(servlet: ZMQServlet, num: int):
    msgs = []
    while len(msgs) < num:
        try:
            msgs.append(await asyncio.wait_for(servlet.recv_msg(servlet.zmq_ins[num]), RECV_TIMEOUT))
            logger.success(msgs[-1])
        except asyncio.TimeoutError:
            logger.error('timed out')
            continue
    return msgs


async def recv_logs(servlet: WALServlet, num: int):
    msgs = []
    while len(msgs) < num:
        try:
            msgs.append(await asyncio.wait_for(servlet.recv_log_msg(), RECV_TIMEOUT))
        except asyncio.TimeoutError:
            continue
    return msgs


async def send_replay_request(servlet: WALServlet, offset: int, total_num: int):
    await servlet.send_log_msg(servlet.get_replay_request(override_offset=offset), 0)
    await recv_logs(servlet, total_num)


async def send_logs(servlet: WALServlet, log: List['core_pb2.LogEntry'], replica: int):
    for msg in log:
        await servlet.send_log_msg(msg, replica)


async def send_requests(servlet: ZMQServlet, requests: List['core_pb2.Request']):
    for msg in requests:
        await servlet.send_msg(msg)



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from typing import Optional, List

from pinecone.specs import Spec
from pinecone import Service
from pinecone.utils import validate_dns_name

from loguru import logger


class TrafficRouter(Spec):

    def __init__(self, name: str, services: List[str], active: Optional[str]):
        """
        :param port: Port to run the gRPC gateway on
        :param native: Whether to run with native python processes or kubernetes
        """
        if active:
            assert active in services
        self._name = name
        self.active_service = active
        self.all_services = services

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str):
        validate_dns_name(value)
        self._name = value

    def set_active(self, service: str):
        assert service in self.all_services
        self.active_service = service

    def add_service(self, service: str):
        if service in self.all_services:
            logger.error("Service already in traffic router")
            return
        self.all_services.append(service)

    def remove_service(self, service: Service):
        self.all_services.remove(service)
        if self.active_service == service:
            self.active_service = None

    def validate(self):
        """Validates the router.

        Perform the following validations:

        - Router name should be DNS compatible
        """
        validate_dns_name(self._name)

    @classmethod
    def from_obj(cls, obj: dict) -> "TrafficRouter":
        assert obj['kind'] == 'TrafficRouter'
        metadata = obj['metadata']
        spec = obj['spec']
        logger.debug(str(spec))

        router = cls(metadata['name'], spec['services'], spec['active'])
        return router

    def to_obj(self) -> dict:
        metadata = {'name': self.name}
        spec = {
            'active': self.active_service,
            'services': self.all_services
        }
        return {
            'version': 'pinecone/v1alpha1',
            'kind': 'TrafficRouter',
            'metadata': metadata,
            'spec': spec
        }



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from pinecone.protos import core_pb2
from pinecone.network.selector import SocketSelector
from pinecone.network.selector.all import AllSelector

import zmq
import zmq.asyncio
import enum
from typing import Callable, Awaitable, Optional, Dict, List


class SocketType(enum.Enum):
    PUB = 1
    SUB = 2
    PAIR = 3
    PULL = 4
    PUSH = 5

    @classmethod
    def proxy_pair(cls, socket_type):
        return {
            cls.PUSH: cls.PULL,
            cls.PULL: cls.PUSH
        }[socket_type]


class Socket(object):

    def __init__(self,
                 bind: bool,
                 sock_type: SocketType,
                 port: int = None,
                 host: str = None):
        self.host = host
        self.port = port
        self.sock_type = sock_type
        self.bind = bind

    def zmq(self, context: 'zmq.context') -> 'zmq.Socket':
        assert (self.port and self.bind) or self.host
        return context.socket({
                                  SocketType.PUB: lambda: zmq.PUB,
                                  SocketType.SUB: lambda: zmq.SUB,
                                  SocketType.PAIR: lambda: zmq.PAIR,
                                  SocketType.PULL: lambda: zmq.PULL,
                                  SocketType.PUSH: lambda: zmq.PUSH
                              }[self.sock_type]())


class ServletSpec:

    _default_socket_selector = AllSelector()

    def __init__(self, handle_msg: Optional[Callable[['core_pb2.Request'], Awaitable['core_pb2.Request']]],
                 in_sockets: List[Socket],
                 out_sockets: Dict[str, List[Socket]],
                 out_socket_selector: 'SocketSelector' = None,
                 native: bool = True,
                 function_name: str = 'no_function',
                 shard: int = 0,
                 replica: Optional[int] = 0,
                 get_stats: Callable = dict,
                 service_name: str = None):
        """
                :param handle_msg:
                :param in_sockets:
                :param out_sockets:
                :param out_socket_selectors: For each path, a function that takes a Request, and number of out_sockets,
                and returns the Request to send to each socket_spec index
                :param native:
                :param function_name:
                :param replica:
                :param get_stats:
                :param service_name:
        """
        if out_socket_selector is None:
            self.out_socket_selector = self._default_socket_selector
        else:
            self.out_socket_selector = out_socket_selector

        self.in_sockets = in_sockets
        self.out_sockets = out_sockets
        self.native = native  # run natively in python with IPC networking
        self.handle_msg = handle_msg
        self.function_name = function_name
        self.replica = replica
        self.get_stats = get_stats
        self.service_name = service_name
        self.shard = shard



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from pinecone.protos import core_pb2
from typing import List, Optional

import abc


class SocketSelector(abc.ABC):

    @classmethod
    @abc.abstractmethod
    def select_socket(cls, orig_msg: 'core_pb2.Request', shards: int) -> List[Optional['core_pb2.Request']]:
        """
        Returns a list of messages to be send to each socket.
        If the messages are identical, a single-item list can be returned
        """
        raise NotImplementedError



from typing import NamedTuple, Iterable, Tuple, Iterator, List
import numpy as np

from pinecone.protos import core_pb2
from pinecone.utils import load_numpy
from pinecone.utils.sentry import sentry_decorator as sentry
from .api_controller import ControllerAPI
from .constants import Config
from .grpc import GRPCClient, GRPCClientConfig, Cursor

__all__ = ["UpsertResult", "DeleteResult", "QueryResult", "FetchResult", "InfoResult", "Index"]


class UpsertResult(NamedTuple):
    """Result of an index request."""

    id: str


class DeleteResult(NamedTuple):
    """Result of a delete request."""

    id: str


class QueryResult(NamedTuple):
    """Result of a query request."""

    ids: Iterable[str]
    scores: Iterable[float]
    data: np.ndarray = None


class FetchResult(NamedTuple):
    """Result of a fetch request"""

    id: str
    vector: np.ndarray


class InfoResult(NamedTuple):
    """Metadata of an index."""

    index_size: str


class Index:
    def __init__(self, name: str, **kwargs):
        """Pinecone Index.

        :param name: name of Pinecone index.
        :type name: str
        :param `**kwargs`:
            See below

        :Keyword Arguments:
            * **batch_size** (*int*) --
              the number of items to batch-process. Defaults to 100.
            * **disable_progress_bar** (*bool*) --
              whether to disable progress bar. Defaults to False.
            * **grcp_config** (*dict*) --
              See :class:`pinecone.grpc.GRPCClientConfig` for grpc client related configurations.
        """
        self.name = name
        self.batch_size = kwargs.pop("batch_size", 100)
        self.disable_progress_bar = kwargs.pop("disable_progress_bar", False)
        self.grpc = self._connect(kwargs.pop("grcp_config", {}))

    def _connect(self, grpc_config: dict):
        """Sets up a connection to an index."""
        api = ControllerAPI(host=Config.CONTROLLER_HOST, api_key=Config.API_KEY)
        status = api.get_status(self.name)
        if not status.get("ready"):
            raise ConnectionError

        if self.name not in api.list_services():
            raise RuntimeError("Index '{}' is not found.".format(self.name))

        host = status.get("host") or Config.CONTROLLER_HOST.split("://")[1].split(":")[0]
        port = status.get("port", -1)
        config = GRPCClientConfig._from_dict(grpc_config)
        return GRPCClient(host=host, port=port, api_key=Config.API_KEY, service_name=self.name, **(config._asdict()))

    def _make_cursor_config(self, **kwargs) -> dict:
        """Assembles cursor configurations."""
        config = {"grpc_client": self.grpc, "batch_size": self.batch_size, "disable_progress_bar": self.disable_progress_bar}
        if "batch_size" in kwargs:
            config["batch_size"] = kwargs.pop("batch_size")
        if "disable_progress_bar" in kwargs:
            config["disable_progress_bar"] = kwargs.pop("disable_progress_bar")
        config.update(**kwargs)
        return config

    def _upsert(self, items: Iterable[Tuple[str, np.ndarray]], namespace: str = None, **kwargs) -> Cursor:
        def _create_index_request(item_batch: Iterable[Tuple[str, np.ndarray]]) -> "core_pb2.Request":
            """Creates an index request from the items."""
            path = "write"
            ids_buffer, vectors_buffer = list(zip(*item_batch))
            ids_buffer = list(map(str, ids_buffer))
            return self.grpc.get_index_request(ids=ids_buffer, data=vectors_buffer, namespace=namespace, path=path)

        def _parse_index_response(response) -> List[UpsertResult]:
            """Parses an index response"""
            return [UpsertResult(id=_id) for _id in response.index.ids]

        config = self._make_cursor_config(
            items=items, create_request_fn=_create_index_request, parse_response_fn=_parse_index_response, **kwargs
        )
        return Cursor(**config)

    def _query(
        self,
        queries: Iterable[np.ndarray],
        namespace: str = None,
        top_k: int = 10,
        include_data: bool = False,
        **kwargs
    ) -> Cursor:
        def _create_query_request(query_batch: Iterable[np.ndarray]) -> "core_pb2.Request":
            """Generates a query request."""
            path = "read"
            return self.grpc.get_query_request(
                data=query_batch, top_k=top_k, include_data=include_data, namespace=namespace, path=path
            )

        def _parse_query_response(response) -> List[QueryResult]:
            """Parses a query response"""
            return [
                QueryResult(
                    ids=matches.ids,
                    scores=matches.scores,
                    data=load_numpy(matches.data) if response.query.include_data else None,
                )
                for matches in response.query.matches
            ]

        config = self._make_cursor_config(
            items=queries, create_request_fn=_create_query_request, parse_response_fn=_parse_query_response, **kwargs
        )
        return Cursor(**config)

    def _fetch(self, ids: Iterable[str], namespace: str = None, **kwargs) -> Cursor:
        def _create_fetch_request(id_batch: Iterable[str]) -> "core_pb2.Request":
            """Generates a delete request."""
            path = "read"
            return self.grpc.get_fetch_request(ids=id_batch, namespace=namespace, path=path)

        def _parse_fetch_response(response) -> List[FetchResult]:
            return [
                FetchResult(id=_id, vector=load_numpy(vector))
                for _id, vector in zip(response.fetch.ids, response.fetch.vectors)
            ]

        config = self._make_cursor_config(
            items=ids, create_request_fn=_create_fetch_request, parse_response_fn=_parse_fetch_response, **kwargs
        )
        return Cursor(**config)

    def _delete(self, ids: Iterable[str], namespace: str = None, **kwargs) -> Cursor:
        def _create_delete_request(id_batch: Iterable[str]) -> "core_pb2.Request":
            """Generates a delete request."""
            path = "write"
            return self.grpc.get_delete_request(ids=id_batch, namespace=namespace, path=path)

        def _parse_delete_response(response) -> List[DeleteResult]:
            return [DeleteResult(id=_id) for _id in response.delete.ids]

        config = self._make_cursor_config(
            items=ids, create_request_fn=_create_delete_request, parse_response_fn=_parse_delete_response, **kwargs
        )
        return Cursor(**config)

    @sentry
    def upsert(
        self, items: Iterable[Tuple[str, np.ndarray]], namespace: str = None, **kwargs
    ) -> Iterable[UpsertResult]:
        """Inserts or updates items.

        An item is an `(id, vector)` tuple.

        Insert an item into the index, or update the item's value by the item's id.

        :param items: tuples of the form (id, numpy.ndarray). The length of an id is limited to **64** characters.
        :type items: Iterable[Tuple[str, np.ndarray]]
        :param namespace: a partition in an index. Defaults to None.
        :type namespace: str, optional
        :param `**kwargs`:
            See below

        :Keyword Arguments:
            * **batch_size** (*int*) --
              overrides the number of items to batch-process.
            * **disable_progress_bar** (*bool*) --
              overrides whether to disable the progress bar
        """
        return self._upsert(items=items, namespace=namespace, **kwargs).collect()

    @sentry
    def unary_upsert(self, item: Tuple[str, np.ndarray], namespace: str = None, **kwargs) -> UpsertResult:
        """Upserts one item at a time.

        This method sends a unary gRPC request.
        See :py:meth:`Index.upsert` for details about the parameters.
        """
        results = self._upsert(items=[item], namespace=namespace, unary=True, disable_progress_bar=True, **kwargs).collect()
        return next(iter(results))

    @sentry
    def stream_upsert(
        self, items: Iterable[Tuple[str, np.ndarray]], namespace: str = None, **kwargs
    ) -> Iterator[UpsertResult]:
        """Upserts items as a stream.

        This method streams the gRPC requests.
        See :py:meth:`Index.upsert` for details about the parameters.
        """
        return self._upsert(items=items, namespace=namespace, **kwargs).stream()

    @sentry
    def query(
        self,
        queries: Iterable[np.ndarray],
        namespace: str = None,
        top_k: int = 10,
        include_data: bool = False,
        **kwargs
    ) -> Iterable[QueryResult]:
        """Sends queries to the index and returns the top results ordered by their scores.

        :param queries: queries
        :type queries: Iterable[np.ndarray]
        :param namespace: a partition in an index. Use default namespace when not specified.
        :type namespace: str, optional
        :param top_k: defaults to 10, the number of top query results to return for each query,
            ordered by their scores
        :type top_k: int, optional
        :param `**kwargs`:
            See below

        :Keyword Arguments:
            * **batch_size** (*int*) --
              overrides the number of items to batch-process.
            * **disable_progress_bar** (*bool*) --
              overrides whether to disable the progress bar
        """
        return self._query(
            queries=queries, namespace=namespace, top_k=top_k, include_data=include_data, **kwargs
        ).collect()

    @sentry
    def unary_query(
        self, query: np.ndarray, namespace: str = None, top_k: int = 10, include_data: bool = False, **kwargs
    ) -> UpsertResult:
        """Sends one query at a time.

        This method sends a unary gRPC request.
        See :py:meth:`Index.query` for details about the parameters.
        """
        results = self._query(
            queries=[query],
            namespace=namespace,
            top_k=top_k,
            include_data=include_data,
            unary=True,
            disable_progress_bar=True,
            **kwargs
        ).collect()
        return next(iter(results))

    @sentry
    def stream_query(
        self,
        queries: Iterable[np.ndarray],
        namespace: str = None,
        top_k: int = 10,
        include_data: bool = False,
        **kwargs
    ) -> Iterator[UpsertResult]:
        """Sends queries as a stream.

        This method streams the gRPC requests.
        See :py:meth:`Index.query` for details about the parameters.
        """
        return self._query(
            queries=queries, namespace=namespace, top_k=top_k, include_data=include_data, **kwargs
        ).stream()

    @sentry
    def fetch(self, ids: Iterable[str], namespace: str = None, **kwargs) -> Iterable[FetchResult]:
        """Fetches items by their ids.

        :param ids: ids of items
        :type ids: Iterable[str]
        :param namespace: a partition in an index. Use default namespace when not specified.
        :type namespace: str, optional
        :param `**kwargs`:
            See below

        :Keyword Arguments:
            * **batch_size** (*int*) --
              overrides the number of items to batch-process.
            * **disable_progress_bar** (*bool*) --
              overrides whether to disable the progress bar
        """
        return self._fetch(ids=ids, namespace=namespace, **kwargs).collect()

    @sentry
    def unary_fetch(self, id: str, namespace: str = None, **kwargs) -> FetchResult:
        """Fetches one item at a time.

        This method sends a unary gRPC request.
        See :py:meth:`Index.fetch` for details about the parameters.
        """
        results = self._fetch(ids=[id], namespace=namespace, unary=True, disable_progress_bar=True, **kwargs).collect()
        return next(iter(results))

    @sentry
    def stream_fetch(self, ids: Iterable[str], namespace: str = None, **kwargs) -> Iterator[FetchResult]:
        """Fetches items as a stream.

        This method streams gRPC requests.
        See :py:meth:`Index.fetch` for details about the parameters.
        """
        return self._fetch(ids=ids, namespace=namespace, **kwargs).stream()

    @sentry
    def delete(self, ids: Iterable[str], namespace: str = None, **kwargs) -> Iterable[DeleteResult]:
        """Deletes items by their ids.

        :param ids: ids of items
        :type ids: Iterable[str]
        :param namespace: a partition in an index. Use default namespace when not specified.
        :type namespace: str, optional
        :param `**kwargs`:
            See below

        :Keyword Arguments:
            * **batch_size** (*int*) --
              overrides the number of items to batch-process.
            * **disable_progress_bar** (*bool*) --
              overrides whether to disable the progress bar
        """
        return self._delete(ids=ids, namespace=namespace, **kwargs).collect()

    @sentry
    def unary_delete(self, id: str, namespace: str = None, **kwargs) -> DeleteResult:
        """Deletes one item at a time.

        This method sends a unary gRPC request.
        See :py:meth:`Index.delete` for details about the parameters.
        """
        results = self._delete(ids=[id], namespace=namespace, unary=True, disable_progress_bar=True, **kwargs).collect()
        return next(iter(results))

    @sentry
    def stream_delete(self, ids: Iterable[str], namespace: str = None, **kwargs) -> Iterator[DeleteResult]:
        """Deletes items as a stream.

        This method streams gRPC requests.
        See :py:meth:`Index.delete` for details about the parameters.
        """
        return self._delete(ids=ids, namespace=namespace, **kwargs).stream()

    @sentry
    def info(self, namespace: str = None) -> InfoResult:
        """Returns information about the index."""

        def _create_info_request(*args) -> "core_pb2.Request":
            path = "write"
            return self.grpc.get_info_request(namespace=namespace, path=path)

        def _parse_info_response(response) -> List[InfoResult]:
            return [InfoResult(index_size=response.info.index_size)]

        config = self._make_cursor_config(
            items=[1],
            create_request_fn=_create_info_request,
            parse_response_fn=_parse_info_response,
            unary=True,
            disable_progress_bar=True,
        )
        return next(iter(Cursor(**config).collect()))

    @sentry
    def close(self):
        """Closes the connection to the index."""

        self.grpc.channel.close()



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from pinecone.functions.model.transformer import Transformer
from pinecone.functions.model.hub import HubModel
from pinecone.utils import dump_numpy, load_numpy
from pinecone.protos import core_pb2

import numpy as np


class HubTransformer (HubModel, Transformer):

    async def transform_vectors(self, vectors: np.ndarray, timeout: float = 30) -> np.ndarray:
        query = core_pb2.QueryRequest(data=dump_numpy(vectors))
        response = await self.call_remote(core_pb2.Request(query=query, path='read'), timeout=timeout)
        return load_numpy(response.query.data)



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from pinecone.functions.model.scorer import Scorer
from pinecone.functions.model.hub import HubModel
from pinecone.utils import dump_numpy
from pinecone.protos import core_pb2

import numpy as np

from typing import Iterable


class HubScorer(HubModel, Scorer):

    async def score(self,
                    data_queries: Iterable[np.ndarray],
                    data_items: Iterable[Iterable[np.ndarray]]
                    ) -> Iterable[Iterable[float]]:

        request = self.build_request(data_queries, data_items)
        response = await self.call_remote(request)
        return [matches.scores for matches in response.query.matches]

    @staticmethod
    def build_request(data_queries, data_items):
        matches = [core_pb2.ScoredResults(
            data=dump_numpy(np.array(items))) for items in data_items]
        request = core_pb2.QueryRequest(data=dump_numpy(np.array(data_queries)), matches=matches)
        return core_pb2.Request(query=request, path='read')



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from pinecone.network.zmq import Socket, SocketType
from pinecone.network.zmq.spec import ServletSpec
from pinecone.utils import constants
from pinecone.protos import core_pb2

import docker
from loguru import logger

import asyncio
import time
import abc


class HubModel(abc.ABC):

    def __init__(self, image: str, docker_host: str = 'localhost', config: dict = None, **kwargs):
        self._image = image
        self.docker_client = None
        self.container = None
        self.servlet = None
        self.docker_host = docker_host
        self.executor = None
        self.log_task = None
        self.dns_tasks = []
        self.config = config or {}
        super().__init__(**kwargs)
        self.config['image'] = image

    @property
    def image(self):
        return self._image

    async def call_remote(self, msg: 'core_pb2.Request', timeout: float = 30) -> 'core_pb2.Request':
        if self.servlet is None:
            logger.error('Must start model using `with` syntax before calling remote function')
            raise Exception
        loop = asyncio.get_event_loop()

        loop.create_task(self.servlet.send_msg(msg))

        logger.warning('awaiting message')
        try:
            response = await asyncio.wait_for(self.servlet.recv_msg(self.servlet.zmq_ins[0]), timeout=timeout)
            return response
        except asyncio.TimeoutError:
            logger.error("CONTAINER TIMED OUT. LOGS:")
            logger.error(self.container.logs())
            raise RuntimeError("Model timed out. See more details above. Consider increasing the timeout argument.")

    async def get_container_logs(self):
        while True:
            await asyncio.sleep(0.5)
            if self.container.status not in ['running', 'created']:
                logger.warning(self.container.logs().decode())
                logger.warning('Model container exited, status {status}', status=self.container.status)
                break

    def image_uri(self):
        return self.image

    def start(self):
        from pinecone.network.zmq.servlet import ZMQServlet

        if self.container and self.container.status == 'running':
            logger.warning('The container is already running.')

        in_sockets = [Socket(False, SocketType.PULL, port=constants.ZMQ_SECONDARY_PORT, host=self.docker_host)]
        out_sockets = [Socket(False, SocketType.PUSH, port=constants.ZMQ_PORT_IN, host=self.docker_host)]
        servlet_spec = ServletSpec(handle_msg=None, in_sockets=in_sockets, out_sockets={'*': out_sockets}, native=False)

        self.servlet = ZMQServlet(servlet_spec)

        ports = {constants.ZMQ_PORT_IN: constants.ZMQ_PORT_IN, constants.ZMQ_SECONDARY_PORT: constants.ZMQ_SECONDARY_PORT}
        command = ['python3', '-m', 'pinecone.functions.model', *self.to_args()]
        self.docker_client = docker.from_env()
        self.container = self.docker_client.containers.run(self.image_uri(), command, ports=ports, detach=True,
                                                           environment=['PYTHONUNBUFFERED==1'])
        time.sleep(1)
        loop = asyncio.get_event_loop()
        self.log_task = loop.create_task(self.get_container_logs())
        self.dns_tasks = self.servlet.refresh_dns()


    def stop(self):
        if self.container:
            self.container.stop()
            self.container = None
        if self.log_task:
            self.log_task.cancel()
        for task in self.dns_tasks:
            task.cancel()

    def __enter__(self):
        self.start()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from typing import NamedTuple, Iterable
import numpy as np

from pinecone.utils import load_numpy, dump_numpy
from pinecone.protos import core_pb2

from pinecone.functions.model import ModelFunction


class QueryResult(NamedTuple):
    ids: Iterable[str] = None
    scores: Iterable[float] = None
    data: np.ndarray = None


class Postprocessor(ModelFunction):
    def transform(self, queries: np.ndarray, matches: Iterable[QueryResult]) -> Iterable[QueryResult]:
        """Transforms the scored matches.
        We only overwrite the original matches with fields in the transformed matches that are not None.
        """
        raise NotImplementedError

    def handle_msg(self, msg: "core_pb2.Request") -> "core_pb2.Request":
        req_type = msg.WhichOneof("body")
        if req_type == "query":
            queries = load_numpy(msg.query.data)
            matches = [
                QueryResult(
                    ids=match.ids, scores=match.scores, data=load_numpy(match.data) if msg.query.include_data else None
                )
                for match in msg.query.matches
            ]
            transformed_matches = self.transform(queries, matches)
            for match, match_tf in zip(msg.query.matches, transformed_matches):
                if match_tf.ids is not None:
                    match.ids[:] = match_tf.ids
                if match_tf.scores is not None:
                    match.scores[:] = match_tf.scores
                if match_tf.data is not None:
                    match.data.CopyFrom(dump_numpy(match_tf.data))
        return msg



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from pinecone.utils.tracing import init_tracer, init_profiler
init_tracer()
init_profiler()

import argparse
from pinecone.network.zmq_service import ZMQService

if __name__ == '__main__':

    parser = argparse.ArgumentParser(description='Load functions with config')

    ZMQService.add_args(parser)

    args, unk = parser.parse_known_args()
    ZMQService.start_with_args(args)



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from pinecone.utils.tracing import init_tracer, init_profiler
init_tracer()
init_profiler()

from pinecone.network.zmq import Socket, SocketType, entrypoints
from pinecone.network.zmq.spec import ServletSpec
from pinecone.utils import load_hub_service, constants
from pinecone import functions

import argparse
import importlib
import json
import asyncio

import concurrent.futures
from functools import partial

if __name__ == '__main__':

    parser = argparse.ArgumentParser(description='Load functions with config')
    functions.Function.add_args(parser)

    args, unk = parser.parse_known_args()

    model = importlib.import_module('model')

    custom_model_cls = load_hub_service(model)
    transformer_service = custom_model_cls(**json.loads(args.config))

    executor = concurrent.futures.ThreadPoolExecutor()

    async def call_handler(msg):
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(executor, partial(transformer_service.handle_msg, msg))

    in_sockets = [Socket(True, SocketType.PULL, port=constants.ZMQ_PORT_IN)]
    out_sockets = [Socket(True, SocketType.PUSH, port=constants.ZMQ_SECONDARY_PORT)]
    spec = ServletSpec(handle_msg=call_handler, in_sockets=in_sockets, out_sockets={'*': out_sockets}, native=False,
                       service_name=transformer_service.name)
    entrypoints.start(spec)



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#
import os
import configparser
from loguru import logger
import sys
from pinecone.specs.service import Service  # noqa
from pinecone.specs.traffic_router import TrafficRouter  # noqa
from pinecone.utils.sentry import sentry_decorator as sentry
from .constants import Config
from .manage import create_index, delete_index, describe_index, list_indexes, ResourceDescription
from .index import Index, UpsertResult, DeleteResult, QueryResult, FetchResult, InfoResult

__all__ = [
    "init",
    "create_index",
    "delete_index",
    "describe_index",
    "list_indexes",
    "ResourceDescription",
    "Index",
    "UpsertResult",
    "DeleteResult",
    "QueryResult",
    "FetchResult",
    "InfoResult",
]

logging_level = os.environ.get("PINECONE_LOGGING", default="ERROR")
logger.remove()
logger.add(sys.stdout, level=logging_level)

__version__ = open(os.path.join(os.path.dirname(__file__), "__version__")).read().strip()


@sentry
def init(api_key: str = None, host: str = None, environment: str = None, config: str = "~/.pinecone", **kwargs):
    """Initializes the Pinecone client.

    :param api_key: Required if not set in config file or by environment variable ``PINECONE_API_KEY``.
    :param host: Optional. Controller host.
    :param environment: Optional. Deployment environment.
    :param config: Optional. An INI configuration file.
    """
    Config.reset(api_key=api_key, controller_host=host, environment=environment, config_file=config, **kwargs)
    if not bool(Config.API_KEY):
        logger.warning("API key is required.")


# Init
init()


#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from pinecone.functions import Function
from pinecone.protos import core_pb2
from pinecone.utils import load_numpy, dump_numpy

from typing import Iterable, List
from collections import defaultdict
from abc import abstractmethod
import numpy as np


class Ranker (Function):

    """
    Encapsulates aggregating and ranking logic in the execution plane
    """

    def __init__(self, *args, num_shards: int = 1, **kwargs):
        super().__init__(*args, **kwargs)
        self.num_shards = num_shards
        self.config['num_shards'] = num_shards
        self.msgs = dict()
        self.shard_counts = defaultdict(set)

    @abstractmethod
    def score(self, q: np.ndarray, vectors: np.ndarray, prev_scores: Iterable[float]) -> List[float]:
        """
        :param q:
        :param vectors:
        :param prev_scores: Existing scores if present
        :return:
        """
        raise NotImplementedError

    def top_k_idx(self, scores: Iterable[float], k: int) -> Iterable[int]:
        topk_idx = np.argpartition(scores, -k)[-k:]
        sorted_topk_idx = topk_idx[np.array(scores)[topk_idx].argsort()[::-1]]
        return sorted_topk_idx

    @staticmethod
    def merge_fetch_msg(orig: 'core_pb2.Request', msg: 'core_pb2.Request') -> 'core_pb2.Request':
        for i, vector in enumerate(msg.fetch.vectors):
            if sum(vector.shape) > 0:
                orig.fetch.vectors[i].CopyFrom(vector)
        return orig

    @staticmethod
    def merge_nd_arrays(orig: 'core_pb2.NdArray', new: 'core_pb2.NdArray'):
        orig_data = load_numpy(orig)
        new_data = load_numpy(new)
        if len(new_data) == 0:
            pass
        elif len(orig_data) == 0:
            orig.CopyFrom(dump_numpy(new_data))
        else:
            orig.CopyFrom(dump_numpy(
                np.concatenate((orig_data, new_data))))
    @staticmethod
    def merge_query_msg(orig: 'core_pb2.Request', msg: 'core_pb2.Request') -> 'core_pb2.Request':
        for orig_matches, new_matches in zip(orig.query.matches, msg.query.matches):
            orig_matches.ids.extend(new_matches.ids)
            orig_matches.scores.extend(new_matches.scores)
            if orig.query.include_data:
                Ranker.merge_nd_arrays(orig_matches.data, new_matches.data)

        orig.routes.append(msg.routes.pop(-1))
        return orig

    @staticmethod
    def merge_info_msg(orig: 'core_pb2.Request', msg: 'core_pb2.Request') -> 'core_pb2.Request':
        orig.info.index_size += msg.info.index_size
        return orig

    @staticmethod
    def merge_delete_msg(orig: 'core_pb2.Request', msg: 'core_pb2.Request') -> 'core_pb2.Request':
        orig.delete.ids.extend(msg.delete.ids)
        return orig

    @staticmethod
    def merge_index_msg(orig: 'core_pb2.Request', msg: 'core_pb2.Request') -> 'core_pb2.Request':
        orig.index.ids.extend(msg.index.ids)
        return orig

    def rank_query_msg(self, query: 'core_pb2.QueryRequest'):
        k = query.top_k
        queries_vectors = load_numpy(query.data)
        for matches, q in zip(query.matches, queries_vectors):
            q = q.reshape(1,-1)
            if len(matches.ids) == 0:  # nothing to rank
                continue
            embeddings = load_numpy(matches.data)
            scores = self.score(q, embeddings, matches.scores)
            k = min(k, len(scores))  # for k>n case
            topk_idx = self.top_k_idx(scores, k)
            sorted_ids = [matches.ids[int(i)] for i in topk_idx]
            sorted_scores = [scores[int(i)] for i in topk_idx]
            if len(embeddings) > 0 and query.include_data:
                sorted_embeddings = dump_numpy(embeddings[topk_idx, :])
                matches.CopyFrom(core_pb2.ScoredResults(
                    ids=sorted_ids, scores=sorted_scores, data=sorted_embeddings))
            else:
                matches.CopyFrom(core_pb2.ScoredResults(
                    ids=sorted_ids, scores=sorted_scores))

    def merge_msg(self, msg: 'core_pb2.Request'):
        request_type = msg.WhichOneof('body')
        if self.msgs.get(msg.request_id) is None:
            self.msgs[msg.request_id] = msg
        elif request_type == 'query':
            self.msgs[msg.request_id] = Ranker.merge_query_msg(
                self.msgs[msg.request_id], msg)
        elif request_type == 'fetch':
            self.msgs[msg.request_id] = Ranker.merge_fetch_msg(
                self.msgs[msg.request_id], msg)
        elif request_type == 'info':
            self.msgs[msg.request_id] = Ranker.merge_info_msg(
                self.msgs[msg.request_id], msg)
        elif request_type == 'index':
            self.msgs[msg.request_id] = Ranker.merge_index_msg(
                self.msgs[msg.request_id], msg)
        elif request_type == 'delete':
            self.msgs[msg.request_id] = Ranker.merge_delete_msg(
                self.msgs[msg.request_id], msg)

    def handle_msg(self, msg: 'core_pb2.Request') -> 'core_pb2.Request':
        if msg.shard_num in self.shard_counts[msg.request_id]:
            return
        self.shard_counts[msg.request_id].add(msg.shard_num)
        self.merge_msg(msg)
        if len(self.shard_counts[msg.request_id]) == self.num_shards:
            result = self.msgs[msg.request_id]
            del self.msgs[msg.request_id]
            del self.shard_counts[msg.request_id]
            if msg.WhichOneof('body') == 'query':
                self.rank_query_msg(result.query)
            return result



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from .exceptions import InvalidOffsetException
from pinecone.utils import replica_kube_hostname, replica_name
from pinecone.protos import core_pb2
from pinecone.network.zmq.servlet import ZMQServlet
from pinecone.utils import open_or_create, get_hostname, wal
from pinecone.utils.constants import MAX_CLIENTS, SEND_TIMEOUT, ZMQ_LOG_PORT, MAX_MSGS_PENDING, RECV_TIMEOUT
from pinecone.network.zmq.spec import ServletSpec, SocketType, Socket

from typing import Iterable, Tuple, List, Union
from loguru import logger

import concurrent.futures
from threading import Lock
import asyncio
import os
import functools
import mmap
import zmq


class WALServlet(ZMQServlet):
    """
    ZMQServlet for distributed stateful functions to provide
    - client ordering guarantees
    - replay recovery from disk
    - leader-follower replication
    """

    def __init__(self, servlet_spec: ServletSpec, num_replicas: int, path: str, volume_request: int = 15):
        super().__init__(servlet_spec)
        self.replica = servlet_spec.replica
        self.num_replicas = num_replicas
        self.wal_size_limit = volume_request * (10 ** 9) * 0.33  # translate to GB

        self.executor = concurrent.futures.ThreadPoolExecutor()

        self.log_in = self.init_socket(self.log_in_socket)
        self.log_outs = [self.init_socket(log_out_socket, disable_loadbalance=True)
                         for log_out_socket in self.log_out_sockets()]

        self.in_buffer = asyncio.Queue()
        self.in_msgs = asyncio.Queue(maxsize=MAX_MSGS_PENDING)

        os.makedirs(path, exist_ok=True)

        self.log_path = os.path.join(path, 'log')
        self.tmp_log_path = os.path.join(path, 'tmp_log')  # for compaction
        self.log_file = open_or_create(self.log_path)
        self.log_lock = Lock()

        self.log_compaction_started = False

        self.last = core_pb2.Request()
        self.size = self.load_size()

        offset_path = os.path.join(path, 'offsets')
        self.offset_file = open_or_create(offset_path, truncate=wal.OFFSET_BYTES*self.num_replicas)
        self.offsets = mmap.mmap(self.offset_file.fileno(), 0)
        self.offset_locks = [Lock() for _ in range(0, self.num_replicas)]

        client_offsets_path = os.path.join(path, 'client_offsets')
        self.client_offsets_file = open_or_create(client_offsets_path, truncate=wal.OFFSET_BYTES*MAX_CLIENTS)

        self.client_offsets = mmap.mmap(self.client_offsets_file.fileno(), 0)
        self.client_offsets_locks = [Lock() for _ in range(0, MAX_CLIENTS)]

    @property
    def log_in_socket(self):
        return Socket(True, SocketType.PULL, ZMQ_LOG_PORT, host=get_hostname())

    def log_out_sockets(self):
        return [
            Socket(False, SocketType.PUSH,
                   host=replica_name(self.spec.function_name, self.spec.shard, replica_id=replica_id) if self.spec.native
                   else replica_kube_hostname(self.spec.function_name, self.spec.shard, replica_id),
                   port=ZMQ_LOG_PORT)
            for replica_id in range(0, self.num_replicas)]

    def load_size(self):
        size = 0
        for pos, entry in self.sync_replay_log(0):
            size += 1
            self.last = entry
        return size

    def cleanup(self):
        self.offset_file.close()
        self.log_file.close()

    @property
    def leader(self):
        return self.replica == 0

    async def poll_log_socks(self):
        while True:
            msg = await self.recv_log_msg()
            await self.handle_log_msg(msg)

    async def poll_sock(self, sock: zmq.Socket):
        loop = asyncio.get_event_loop()
        while True:
            msg = await self.recv_msg(sock)
            if self.use_wal(msg):
                await self.in_msgs.put(msg)
            else:
                loop.create_task(self.handle_msg(msg))

    def start_polling(self) -> List[asyncio.Task]:
        loop = asyncio.get_event_loop()
        return [*(loop.create_task(self.poll_sock(sock)) for sock in self.zmq_ins), loop.create_task(self.poll_log_socks())]

    def use_wal(self, msg: core_pb2.Request) -> bool:
        req_type = msg.WhichOneof('body')
        return req_type in ['index', 'delete']

    async def handle_msg_wal(self, msg: core_pb2.Request):
        new_offset = await self.put(msg)
        client_id = msg.client_id
        client_offset = msg.client_offset
        await self.handle_msg(msg)
        await self.set_client_offset(client_id, client_offset + 1)
        await self.ack(new_offset)

    async def handle_log_msg(self, msg: 'core_pb2.LogEntry'):
        if not self.leader:
            try:
                if await self.put(msg.entry, offset=msg.offset):
                    await self.handle_msg(msg.entry)
            except InvalidOffsetException:
                logger.warning(f"Received invalid offset {msg.offset}")
                await self.send_log_msg(self.get_replay_request(), 0)
        else:
            await self.remote_ack(msg.offset, msg.ack.replica, msg.ack.replay)
            del msg

    async def consume_msgs(self):
        while True:
            msg = await self.in_msgs.get()
            if self.is_next(msg.client_id, msg.client_offset):
                await self.handle_msg_wal(msg)
                while True:
                    try:
                        await self.in_msgs.put(self.in_buffer.get_nowait())
                    except asyncio.QueueEmpty:
                        break
            else:
                # logger.error(f"Invalid client_offset {msg.client_offset}")
                if msg.client_offset >= self.get_client_offset(msg.client_id):
                    await self.in_buffer.put(msg)
                # elif msg.client_offset == self.get_client_offset(msg.client_id) - 1:
                #     await self.send_msg(msg)

    async def recv_log_msg(self):
        msg = await self.log_in.recv()
        msg_pb = core_pb2.LogEntry()
        msg_pb.ParseFromString(msg)
        return msg_pb

    async def send_log_msg(self, msg: Union['core_pb2.LogEntry', bytes], replica: int):
        await self.log_outs[replica].send(msg if type(msg) == bytes else msg.SerializeToString())

    async def run_sync(self, func, *args, **kwargs):
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(self.executor, functools.partial(func, *args, **kwargs))

    def get_offset(self, replica: int):
        start = wal.OFFSET_BYTES * replica
        return int.from_bytes(self.offsets[start:start + wal.OFFSET_BYTES], wal.BYTEORDER)

    def offsets_list(self):
        return [self.get_offset(replica) for replica in range(0, self.num_replicas)]

    def write_offset(self, replica: int, offset: int):
        with self.offset_locks[replica]:
            self.offsets.seek(replica * wal.OFFSET_BYTES)
            self.offsets.write(offset.to_bytes(wal.OFFSET_BYTES, wal.BYTEORDER))
            self.offsets.flush()

    async def set_offset(self, replica: int, offset: int):
        await self.run_sync(self.write_offset, replica, offset)

    def get_client_offset(self, client_id: int):
        if client_id >= MAX_CLIENTS:
            return 0
        with self.client_offsets_locks[client_id]:
            start = client_id * wal.OFFSET_BYTES
            return int.from_bytes(self.client_offsets[start:start + wal.OFFSET_BYTES], wal.BYTEORDER)

    def is_next(self, client_id: int, client_offset: int):
        if self.leader:
            if client_offset == 0:
                self.sync_set_client_offset(client_id, 0)  # should make this async
            return client_offset == self.get_client_offset(client_id)
        else:
            return True

    def sync_set_client_offset(self, client_id: int, client_offset: int):
        if client_id >= MAX_CLIENTS:
            return
        with self.client_offsets_locks[client_id]:
            start = client_id * wal.OFFSET_BYTES
            self.client_offsets[start:start + wal.OFFSET_BYTES] = client_offset.to_bytes(wal.OFFSET_BYTES, wal.BYTEORDER)
            self.client_offsets.flush()

    async def set_client_offset(self, client_id: int, client_offset: int):
        if self.get_client_offset(client_id) > client_offset:
            return
        await self.run_sync(self.sync_set_client_offset, client_id, client_offset)

    def sync_put(self, entry: core_pb2.Request, offset: int):
        if (not offset) or (self.size + 1) == offset:
            with self.log_lock:
                wal.write_log_entry(self.log_file, entry)
                self.size += 1
                # we need to copy the message, because it might get changed later on
                self.last.CopyFrom(entry)
                logfile_size = self.log_file.tell()
            return logfile_size, self.size

        elif offset <= self.size:
            return self.log_file.tell(), None
        else:
            raise InvalidOffsetException

    async def put(self, item: core_pb2.Request, offset: int = None):
        if not self.leader and offset is None:
            raise RuntimeError(f"{get_hostname()} Attempted to write directly to replica {self.replica} instead of leader")
        logfile_size, offset = await self.run_sync(self.sync_put, item, offset)
        if logfile_size > self.wal_size_limit and not self.log_compaction_started:
            loop = asyncio.get_event_loop()
            loop.create_task(self.start_compact_log())
        return offset

    def sync_replay_log(self, offset: int, end_offset: int = None) -> Iterable[Tuple[int, core_pb2.Request]]:
        """
        Replay log from offset 0 (inclusive) to end_offset (exclusive)
        :param offset:
        :param end_offset:
        :return:
        """
        with self.log_lock:
            self.log_file.seek(0)
            pos = 0
            while True:
                entry, msg_size = wal.read_entry(self.log_file)

                if msg_size == 0 or (end_offset and pos >= end_offset):
                    break
                if pos >= offset:
                    yield pos, entry
                pos += 1

    def handle_replay_request(self, loop: asyncio.AbstractEventLoop, replica: int, offset: int, end_offset: int = None):
        logger.info(f"Replaying log from offset {offset} to replica {replica}")
        for pos, entry in self.sync_replay_log(offset, end_offset=end_offset):
            asyncio.run_coroutine_threadsafe(self.send_log_msg(core_pb2.LogEntry(entry=entry,
                                                                                 offset=pos), replica), loop)

    def get_leader_log_backlog(self, replica: int) -> Iterable[core_pb2.LogEntry]:
        return [core_pb2.LogEntry(entry=entry, offset=offset)
                for offset, entry in self.sync_replay_log(self.get_offset(replica))]

    def sync_ack(self, offset: int):
        self.write_offset(self.replica, offset)

    async def ack(self, offset: int):
        prev_offset = self.get_offset(self.replica)
        if prev_offset > offset:
            logger.warning('Acked offset less than current offset')
        # elif prev_offset + 1 > offset:
        #     logger.warning('Acked by more than one offset')

        await self.set_offset(self.replica, offset)
        if self.leader:
            msg = core_pb2.LogEntry(entry=self.last, offset=offset)
            serialized_msg = msg.SerializeToString()
            for r in range(1, self.num_replicas):
                await self.send_log_msg(serialized_msg, r)

        else:
            ack = core_pb2.Ack(replica=self.replica)
            await self.send_log_msg(core_pb2.LogEntry(offset=offset, ack=ack), 0)

    async def remote_ack(self, start_offset: int, replica: int, replay=False):
        if replay:
            leader_offset = self.get_offset(self.replica)
            event_loop = asyncio.get_event_loop()
            await self.run_sync(self.handle_replay_request, event_loop, replica, start_offset, end_offset=leader_offset)

    def get_replay_request(self, override_offset=None):
        ack = core_pb2.Ack(replica=self.replica, replay=True)
        offset = override_offset or self.get_offset(self.replica)
        logger.info(f"Sending replay request for offset {offset}")
        return core_pb2.LogEntry(offset=offset, ack=ack)

    def load_sync(self, handler):
        logger.info(f'Recovering from local WAL: offset: {self.get_offset(self.replica)}')
        for offset, entry in self.sync_replay_log(self.get_offset(self.replica)):
            try:
                handler(entry)  # type: core_pb2.Request
            except Exception as e:  # ignore messages that fail
                logger.info(f'Error handling msg while recovering from WAL. offset: {offset} error: {e}')
            self.sync_ack(offset)

    def swap_log_files(self, end_file_offset: int):
        tmp_log_file = open(self.tmp_log_path, "r+b")

        with self.log_lock:
            self.log_file.seek(end_file_offset)
            while True:
                entry, msg_size = wal.read_entry(self.log_file)
                if msg_size == 0:
                    break
                wal.write_log_entry(tmp_log_file, entry)
            tmp_log_file.close()
            self.log_file.close()
            os.remove(self.log_path)
            os.rename(self.tmp_log_path, self.log_path)
            self.log_file = open(self.log_path, "r+b")

    async def start_compact_log(self):
        with self.log_lock:
            self.log_compaction_started = True
            end_file_offset = self.log_file.tell()
            self.log_file.flush()
        loop = asyncio.get_event_loop()
        await loop.run_in_executor(concurrent.futures.ProcessPoolExecutor(), wal.compact_log,
                                   self.log_path, self.tmp_log_path, end_file_offset)

        await self.run_sync(self.swap_log_files, end_file_offset)
        self.log_compaction_started = False

    @property
    def all_socks(self):
        return [*super().all_socks, self.log_in, *self.log_outs]



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from pinecone.api_base import BaseAPI

__all__ = ["HubAPI"]


class HubAPI(BaseAPI):
    """API calls to the hub"""

    def get_login(self):
        """Returns the docker command for logging into the Hub."""
        return self.post("/login")

    def create_repository(self, repository_name):
        return self.post("/repositories",  json={"repository_name": repository_name})

    def delete_repository(self, repository_name):
        return self.delete("/repositories/{}".format(repository_name))

    def list_repository_tags(self, repository_name):
        """List repository tags"""
        return self.get("/repositories/{}/tags".format(repository_name))

    def list_repositories(self):
        """Show all the repositories for the user."""
        return self.get("/repositories")




#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from pinecone.functions.index import Index
from pinecone.protos import core_pb2
from pinecone.utils import load_numpy, dump_numpy, logger
from pinecone.utils.constants import MAX_ID_LENGTH
import numpy as np
import json
from typing import Iterable, Tuple
from abc import abstractmethod


class VectorIndex(Index):
    """
    Generic interface for defining vector index
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @abstractmethod
    def add_data(self, data: np.ndarray, ids: Iterable[str]):
        """
        Add data to an index
        :param data:
        :param vectors: list of vectors to add
        :param ids: list of ids that correspond to those vectors
        """
        raise NotImplementedError

    @abstractmethod
    def delete_data(self, ids: Iterable[str]):
        """
        Delete data points from index by id
        :param ids: ids to delete
        :return:
        """
        raise NotImplementedError

    @abstractmethod
    def query(self, vectors: np.ndarray, k: int, include_data: bool = False) -> (Iterable[Tuple[Iterable[str],
                                                                                 Iterable[float], np.ndarray]]):
        """
        Query an index
        :param include_data: If True, returns data associated with the matched ids.
        :param vectors: Vectors to query
        :param k: number of similar items to return
        :return: list of ids of similar items for each vector
        """
        raise NotImplementedError

    @abstractmethod
    def fetch_data(self, ids: Iterable[str], default: np.ndarray = None) -> Iterable[np.ndarray]:
        """
        Fetch raw vectors from the index
        :param ids: ids to fetch
        :param default: default value for cases where the vector the id wasn't found
        :return: list of vectors associated with the received ids
        """
        raise NotImplementedError

    @abstractmethod
    def size(self) -> int:
        """
        :return: number of items in the index
        """

    @abstractmethod
    def dimension(self) -> int:
        """
        :return: dimension of the indexed vectors
        """

    @staticmethod
    def downcast_data(data: np.ndarray) -> np.ndarray:
        if data.dtype == np.float32:
            return data
        if not np.can_cast(data.dtype, np.float32, casting='same_kind'):
            raise TypeError(f"Data type must be float32 compatible (float64, int64, short, etc.). Given: {data.dtype}")
        return data.astype('float32', casting='same_kind', copy=False)

    def validate_dimension(self, data: np.ndarray):
        if self.dimension() and data.shape[1] != self.dimension():
            raise ValueError(f"Data dimension must be consistent {data.shape[1]} != {self.dimension()}")

    def export_metadata(self):
        logger.opt(raw=True).info(json.dumps({"recordType": "vectorIndex.metadata",
                                              "name": self.name,
                                              "d": self.dimension(),
                                              **self.config}) + "\n")

    @staticmethod
    def validate_ids(ids: Iterable[str]):
        if MAX_ID_LENGTH == 0:
            return  # no limit
        if any(len(id_) > MAX_ID_LENGTH for id_ in ids):
            raise ValueError(f"Vector IDs should be no longer than {MAX_ID_LENGTH} characters")

    def handle_msg(self, msg: 'core_pb2.Request') -> 'core_pb2.Request':
        req_type = msg.WhichOneof('body')
        if req_type == 'delete':
            self.delete_data(msg.delete.ids)
        elif req_type == 'index':
            if self.full:
                raise RuntimeError("Index is full, cannot accept upserts")
            VectorIndex.validate_ids(msg.index.ids)
            data = VectorIndex.downcast_data(load_numpy(msg.index.data))
            if len(data.shape) == 1:
                data = data.reshape(1, -1)
            self.validate_dimension(data)
            self.add_data(data, msg.index.ids)
            msg.index.data.Clear()  # data is not needed along the way
        elif req_type == 'query':
            vectors = VectorIndex.downcast_data(load_numpy(msg.query.data))
            if len(vectors.shape) == 1:
                vectors = vectors.reshape(1, -1)
            self.validate_dimension(vectors)
            all_matches = self.query(vectors, msg.query.top_k, msg.query.include_data)
            for matches in all_matches:
                ids, scores, data = matches
                if msg.query.include_data:
                    msg.query.matches.append(core_pb2.ScoredResults(ids=ids, scores=scores, data=dump_numpy(data)))
                else:
                    msg.query.matches.append(core_pb2.ScoredResults(ids=ids, scores=scores))
        elif req_type == 'info':
            msg.info.index_size += self.size()
        elif req_type == 'fetch':
            default_value = np.array([], dtype=np.float32)
            msg.fetch.vectors.extend(
                [dump_numpy(vector) for vector in self.fetch_data(msg.fetch.ids, default=default_value)])
        return msg



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

import fire
import configparser
import pathlib

__all__ = ["CLI"]


class CLI:
    def init(self, api_key: str = None):
        """Configures the Pinecone client.

        Usage:

        .. code-block:: bash

            pinecone init --api_key=YOUR_API_KEY

        :param api_key: your Pinecone API key
        """
        # Construct config
        config = configparser.ConfigParser()
        config["default"] = {
            "api_key": api_key,
        }
        config["default"] = {key: val for key, val in config["default"].items() if val}
        # Write config file
        with pathlib.Path.home().joinpath(".pinecone").open("w") as configfile:
            config.write(configfile)


def main():
    try:
        fire.Fire(CLI)
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    main()



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from typing import Iterable, Tuple

import numpy as np

from pinecone.functions.index.vector import VectorIndex


class NumpyIndex(VectorIndex):

    def setup(self):
        self._ids = []
        self._vectors = None

    def add_data(self, data: np.ndarray, ids: Iterable[str]):
        self._ids.extend(ids)
        if self._vectors:
            self._vectors = np.concatenate((self._vectors, data))
        else:
            self._vectors = data

    def size(self) -> int:
        return len(self._ids)

    def query(self, vectors: np.ndarray, k: int) -> (Iterable[Tuple[Iterable[str], Iterable[float]]]):
        results = []
        for vector in vectors:
            scores = np.dot(vector, self._vectors).flatten()
            idx = scores.argsort()
            results.append(([self._ids[i] for i in idx][:k], [scores[i] for i in idx][:k]))
        return results



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

from pinecone.functions import Function
from pinecone.runnable import Runnable
from pinecone.network.zmq import Socket, ServletSpec, entrypoints
from pinecone.network.selector.shards import ShardSelector
from pinecone.utils.constants import ZMQ_PORT_IN, ZMQ_LOG_PORT, ZMQ_CONTROL_PORT
from pinecone.utils import load_hub_service, module_name, parse_hostname, get_hostname
from pinecone.protos import core_pb2

import concurrent.futures
import asyncio
import json
from typing import Dict, List, Optional
from pydoc import locate
import os
import random
import importlib
import argparse
import traceback

from loguru import logger
from ddtrace.contrib.asyncio import run_in_executor


class ZMQService(Runnable):

    @classmethod
    def from_args(cls, args):
        """
        :param functions:
        :param args:
        :return:
        """
        # logger.warning("loading {functions}", functions=args.function)
        if args.function in ['pinecone.functions.model.transformer.hub.HubTransformer',
                             'pinecone.functions.model.scorer.hub.HubScorer',
                             'pinecone.functions.model.HubFunction']:

            model = importlib.import_module('model')
            service_cls = load_hub_service(model)
            logger.info('loaded functions class {cls}', cls=service_cls)
            args.config = json.dumps(
                {k: v for k, v in json.loads(args.config).items() if k != 'image'})
        else:
            service_cls = locate(args.function)
        service = service_cls.from_args(args)
        return cls(service, json.loads(args.next), args.service, native=args.native)

    def to_args(self):
        """
        Return the command line args to run this instance
        :return:
        """
        return ['pinecone.functions',
                '--function', module_name(self.function),
                '--next', json.dumps(self.next_services,
                                     separators=(',', ':')),
                '--service', self.service_name,
                *(['--native'] if self.native else []),
                *self.function.to_args()]

    @classmethod
    def add_args(cls, parser: argparse.ArgumentParser):
        parser.add_argument('--function', type=str,
                            help='module of functions class to run')
        parser.add_argument('--next', type=str,
                            help='mapping of next functions in the DAG')
        parser.add_argument('--native', action='store_true')
        parser.add_argument('--service', type=str,
                            help='name of the running runner')
        Function.add_args(parser)

    @classmethod
    def start_with_args(cls, args):
        servlet = cls.from_args(args)
        servlet.start()

    def __init__(self, function: Function, next_services: Dict[str, List[str]], service_name: str, native=True):
        """
        :param service: The functions to mount with a ZMQServlet
        :param next: The next functions in the DAG
        """

        self.function = function
        self.next_services = next_services
        self.native = native
        self.service_name = service_name
        self.executor = None

    def get_sockets(self):
        out_type = self.function.out_sock_type
        in_type = self.function.in_sock_type

        in_sockets = [Socket(True, in_type,
                             host=get_hostname(),
                             port=ZMQ_PORT_IN)]

        out_sockets = {r: [Socket(False, out_type, host=s, port=ZMQ_PORT_IN) for s in n] for r, n in
                       self.next_services.items() if n}

        return in_sockets, out_sockets

    async def proxy_msg(self, msg):
        return msg

    async def run_handler_in_executor(self, msg: 'core_pb2.Request'):
        loop = asyncio.get_event_loop()
        if self.executor is None:
            max_workers = int(os.environ.get('PINECONE_HANDLER_WORKERS', 1))
            self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=max_workers)
        try:
            result = await run_in_executor(loop, self.executor, self.function.handle_msg, msg)
            return result
        except Exception as e:
            msg.status.code = core_pb2.Status.StatusCode.ERROR
            details = core_pb2.Status.Details(function=self.name,
                                              function_id=get_hostname(),
                                              exception=str(e),
                                              traceback=traceback.format_exc())
            details.time.GetCurrentTime()
            msg.status.details.append(details)
            logger.error(details.traceback)
            return msg

    def start(self, hostname=None):
        # FORMAT: <function.name>-s<shard_id>-r<replica_id>]
        if hostname:
            os.environ['HOSTNAME'] = hostname
        hostname = get_hostname()

        if hostname is None:
            raise RuntimeError('Must set HOSTNAME to run 0mq service')

        parsed_shard, parsed_replica = parse_hostname(self.name, hostname)

        # XXX[Fei]: chance of collision?
        replica_id = random.randint(1, 999999) if parsed_replica is None else parsed_replica
        shard_id = 0 if parsed_shard is None else parsed_shard
        logger.debug(f"PARSED SHARD: {shard_id}, PARSED REPLICA: {replica_id} hostname: {hostname}")

        self.function.id = replica_id
        in_sockets, out_sockets = self.get_sockets()

        selector = ShardSelector()

        servlet_spec = ServletSpec(self.run_handler_in_executor, in_sockets, out_sockets, selector, self.native,
                                   self.function.name, shard_id, replica_id,
                                   self.function.get_stats,
                                   service_name=self.service_name)

        self.function.setup()

        entrypoints.start(servlet_spec, volume_request=self.volume_request,
                          persistent_dir=os.path.join(self.function.default_persistent_dir, 'wal'),
                          num_replicas=self.replicas, sync_msg_handler=self.function.handle_msg,
                          stats_exporter=self.function.export_stats)


    def cleanup(self):
        self.function.cleanup()

    @property
    def name(self):
        return self.function.name

    @property
    def image(self):
        return self.function.image

    @property
    def replicas(self) -> int:
        return self.function.replicas

    @property
    def shards(self) -> int:
        return self.function.shards

    @property
    def ports(self) -> List[int]:
        return [ZMQ_PORT_IN, ZMQ_LOG_PORT, ZMQ_CONTROL_PORT]

    @property
    def ext_port(self) -> Optional[int]:
        return

    @property
    def memory_request(self) -> int:
        return self.function.memory_request

    @property
    def cpu_request(self) -> int:
        return self.function.cpu_request

    @property
    def volume_request(self) -> Optional[int]:
        return self.function.volume_request



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

import json
from pinecone.specs.service import Service
from pinecone.specs.traffic_router import TrafficRouter
from typing import Union
mapping = {
    'Service': Service,
    'TrafficRouter': TrafficRouter
}


def load_json_spec(spec_json: str) -> Union[Service, TrafficRouter]:
    spec = json.loads(spec_json)
    return mapping[spec['kind']].from_obj(spec)



#
# Copyright (c) 2020-2021 Pinecone Systems Inc. All right reserved.
#

import yaml
import json
import abc


class Spec:
    """
    Object that can be deployed to a controller
    """

    @property
    @abc.abstractmethod
    def name(self) -> str:
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def from_obj(cls, obj: dict) -> "Spec":
        raise NotImplementedError

    @abc.abstractmethod
    def to_obj(self) -> dict:
        raise NotImplementedError

    def validate(self):
        pass

    @classmethod
    def from_json(cls, spec: str) -> "Spec":
        """Creates class instance from json string."""
        obj = None
        try:
            obj = json.loads(spec)
        except Exception as e:
            raise Exception(f"Unable to load spec: {e}")
        return cls.from_obj(obj)

    def to_json(self) -> str:
        return json.dumps(self.to_obj())

    def to_yaml(self) -> str:
        return yaml.safe_dump(self.to_obj())
